﻿using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Analysis;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis.Symbolic;
using Variable = HeuristicLab.Problems.DataAnalysis.Symbolic.Variable;

namespace HeuristicLab.EvolutionaryTracking.Analyzers.BuildingBlocks {
  [StorableClass]
  class Poly10BuildingBlocksAnalyzer : SingleSuccessorOperator, IAnalyzer {
    private const string SymbolicExpressionTreeParameterName = "SymbolicExpressionTree";
    private const string UpdateIntervalParameterName = "UpdateInterval";
    private const string UpdateCounterParameterName = "UpdateCounter";
    private const string ResultsParameterName = "Results";
    private const string MatchVariablesParameterName = "MatchVariableNames";
    private const string MatchVariableWeightsParameterName = "MatchVariableWeights";
    private const string MatchConstantValuesParameterName = "MatchConstantValues";
    private const string SymbolicExpressionTreeGrammarParameterName = "SymbolicExpressionTreeGrammar";
    private const string ClonedSymbolicExpressionTreeGrammarParameterName = "ClonedSymbolicExpressionTreeGrammar";

    private MaximalCommonSubsequenceCalculator maximalCommonSubsequenceCalculator;

    private List<BuildingBlockGroup> buildingBlockGroups;

    #region Parameters
    public ValueParameter<IntValue> UpdateIntervalParameter {
      get { return (ValueParameter<IntValue>)Parameters[UpdateIntervalParameterName]; }
    }
    public ValueParameter<IntValue> UpdateCounterParameter {
      get { return (ValueParameter<IntValue>)Parameters[UpdateCounterParameterName]; }
    }
    public LookupParameter<ResultCollection> ResultsParameter {
      get { return (LookupParameter<ResultCollection>)Parameters[ResultsParameterName]; }
    }
    public IScopeTreeLookupParameter<ISymbolicExpressionTree> SymbolicExpressionTreeParameter {
      get { return (IScopeTreeLookupParameter<ISymbolicExpressionTree>)Parameters[SymbolicExpressionTreeParameterName]; }
    }
    public ValueParameter<BoolValue> MatchVariableNamesParameter {
      get { return (ValueParameter<BoolValue>)Parameters[MatchVariablesParameterName]; }
    }

    public ValueParameter<BoolValue> MatchVariableWeightsParameter {
      get { return (ValueParameter<BoolValue>)Parameters[MatchVariableWeightsParameterName]; }
    }

    public ValueParameter<BoolValue> MatchConstantValuesParameter {
      get { return (ValueParameter<BoolValue>)Parameters[MatchConstantValuesParameterName]; }
    }
    public IValueLookupParameter<ISymbolicExpressionGrammar> SymbolicExpressionTreeGrammarParameter {
      get { return (IValueLookupParameter<ISymbolicExpressionGrammar>)Parameters[SymbolicExpressionTreeGrammarParameterName]; }
    }
    public ILookupParameter<ISymbolicExpressionGrammar> ClonedSymbolicExpressionTreeGrammarParameter {
      get { return (ILookupParameter<ISymbolicExpressionGrammar>)Parameters[ClonedSymbolicExpressionTreeGrammarParameterName]; }
    }
    #endregion

    #region Parameter properties
    public IntValue UpdateInterval {
      get { return UpdateIntervalParameter.Value; }
    }

    public IntValue UpdateCounter {
      get { return UpdateCounterParameter.Value; }
    }

    public ResultCollection Results {
      get { return ResultsParameter.ActualValue; }
    }
    public ItemArray<ISymbolicExpressionTree> SymbolicExpressionTrees {
      get { return SymbolicExpressionTreeParameter.ActualValue; }
    }
    public ISymbolicExpressionGrammar SymbolicExpressionTreeGrammar {
      get { return ClonedSymbolicExpressionTreeGrammarParameter.ActualValue; }
    }
    #endregion

    [StorableConstructor]
    protected Poly10BuildingBlocksAnalyzer(bool deserializing) : base(deserializing) { }

    public override IDeepCloneable Clone(Cloner cloner) { return new Poly10BuildingBlocksAnalyzer(this, cloner); }

    protected Poly10BuildingBlocksAnalyzer(Poly10BuildingBlocksAnalyzer original, Cloner cloner)
      : base(original, cloner) {
      this.buildingBlockGroups = new List<BuildingBlockGroup>(original.buildingBlockGroups);
      this.maximalCommonSubsequenceCalculator = original.maximalCommonSubsequenceCalculator;
    }

    public Poly10BuildingBlocksAnalyzer() {
      Parameters.Add(new ScopeTreeLookupParameter<ISymbolicExpressionTree>(SymbolicExpressionTreeParameterName, "The symbolic expression trees to analyze."));
      Parameters.Add(new ValueParameter<IntValue>(UpdateIntervalParameterName, "The interval in which the tree length analysis should be applied.", new IntValue(1)));
      Parameters.Add(new ValueParameter<IntValue>(UpdateCounterParameterName, "The value which counts how many times the operator was called since the last update", new IntValue(0)));
      Parameters.Add(new ValueLookupParameter<ResultCollection>(ResultsParameterName, "The results collection where the analysis values should be stored."));
      // matching predicates      
      Parameters.Add(new ValueParameter<BoolValue>(MatchVariablesParameterName, "Specify if the symbolic expression tree comparer should match variable names.", new BoolValue(true)));
      Parameters.Add(new ValueParameter<BoolValue>(MatchVariableWeightsParameterName, "Specify if the symbolic expression tree comparer should match variable weights.", new BoolValue(false)));
      Parameters.Add(new ValueParameter<BoolValue>(MatchConstantValuesParameterName, "Specify if the symbolic expression tree comparer should match constant values.", new BoolValue(false)));
      Parameters.Add(new ValueLookupParameter<ISymbolicExpressionGrammar>(SymbolicExpressionTreeGrammarParameterName, "The tree grammar that defines the correct syntax of symbolic expression trees that should be created."));
      Parameters.Add(new LookupParameter<ISymbolicExpressionGrammar>(ClonedSymbolicExpressionTreeGrammarParameterName, "An immutable clone of the concrete grammar that is actually used to create and manipulate trees."));
      buildingBlockGroups = new List<BuildingBlockGroup>();

      maximalCommonSubsequenceCalculator = new MaximalCommonSubsequenceCalculator();
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      if (maximalCommonSubsequenceCalculator == null)
        maximalCommonSubsequenceCalculator = new MaximalCommonSubsequenceCalculator();

      if (buildingBlockGroups == null)
        buildingBlockGroups = new List<BuildingBlockGroup>();

      buildingBlockGroups.Clear();

      if (!Parameters.ContainsKey(ClonedSymbolicExpressionTreeGrammarParameterName))
        Parameters.Add(new LookupParameter<ISymbolicExpressionGrammar>(ClonedSymbolicExpressionTreeGrammarParameterName, "An immutable clone of the concrete grammar that is actually used to create and manipulate trees."));
    }
    public override void ClearState() {
      buildingBlockGroups.Clear();
      base.ClearState();
    }

    public bool EnabledByDefault { get; private set; }

    public override IOperation Apply() {
      UpdateCounter.Value++;
      if (UpdateCounter.Value != UpdateInterval.Value) return base.Apply();
      UpdateCounter.Value = 0;

      if (ClonedSymbolicExpressionTreeGrammarParameter.ActualValue == null) {
        SymbolicExpressionTreeGrammarParameter.ActualValue.ReadOnly = true;
        IScope globalScope = ExecutionContext.Scope;
        while (globalScope.Parent != null)
          globalScope = globalScope.Parent;

        globalScope.Variables.Add(new Core.Variable(ClonedSymbolicExpressionTreeGrammarParameterName, (ISymbolicExpressionGrammar)SymbolicExpressionTreeGrammarParameter.ActualValue.Clone()));
      }

      if (buildingBlockGroups.Count == 0)
        PrepareFragments();

      var comparer = new SymbolicExpressionTreeNodeSimilarityComparer {
        MatchConstantValues = MatchConstantValuesParameter.Value.Value,
        MatchVariableNames = MatchVariableNamesParameter.Value.Value,
        MatchVariableWeights = MatchVariableWeightsParameter.Value.Value
      };

      maximalCommonSubsequenceCalculator.SimilarityComparer = comparer;

      DataTable dataTable;
      if (!Results.ContainsKey("BuildingBlocksFrequencies")) {
        dataTable = new DataTable("BuildingBlocksFrequencies");
        Results.Add(new Result("BuildingBlocksFrequencies", dataTable));
        foreach (var group in buildingBlockGroups) {
          dataTable.Rows.Add(new DataRow(group.Name) { VisualProperties = { StartIndexZero = true } });
        }
      } else {
        dataTable = (DataTable)Results["BuildingBlocksFrequencies"].Value;
      }

      var canonicalSorter = new SymbolicExpressionTreeCanonicalSorter();
      foreach (var tree in SymbolicExpressionTrees) {
        canonicalSorter.SortSubtrees(tree);
      }

      foreach (var group in buildingBlockGroups) {
        double frequency = 0.0;

        foreach (var tree in SymbolicExpressionTrees) {
          foreach (var bb in group.BuildingBlocks) {
            var maxCommonSequence = maximalCommonSubsequenceCalculator.Calculate(tree.Root, bb.Fragment.Root);
            if (maxCommonSequence.Count == bb.Fragment.Length) {
              frequency++;
              break;
            }
            canonicalSorter.SortSubtrees(bb.Fragment.Root);
          }

          //        var frequency = SymbolicExpressionTrees.Count(x => maximalCommonSubsequenceCalculator.Calculate())
          //        var frequency = SymbolicExpressionTrees.Count(x => x.Root.ContainsFragment(bb.Fragment, comparer));
        }
        double relativeFrequency = frequency / SymbolicExpressionTrees.Length;
        dataTable.Rows[group.Name].Values.Add(relativeFrequency);
      }

      return base.Apply();
    }

    private void PrepareFragments() {
      buildingBlockGroups.Clear();

      var multiplication = new Multiplication();
      var variable = (Variable)SymbolicExpressionTreeGrammar.Symbols.First(s => s.Name.Equals("Variable"));
      var variableNames = variable.VariableNames.ToList();

      {  // (* x1 x2)
        var n = multiplication.CreateTreeNode();
        var x1 = new VariableTreeNode(variable) { VariableName = variableNames[0] };
        var x2 = new VariableTreeNode(variable) { VariableName = variableNames[1] };
        n.AddSubtree(x1);
        n.AddSubtree(x2);
        var bb = new BuildingBlock { Name = "(* x1 x2)", Fragment = new Fragment { Root = n } };
        buildingBlockGroups.Add(new BuildingBlockGroup { Name = "(* x1 x2)", BuildingBlocks = new List<BuildingBlock> { bb } });
      }
      { // (* x3 x4)
        var n = multiplication.CreateTreeNode();
        var x3 = new VariableTreeNode(variable) { VariableName = variableNames[2] };
        var x4 = new VariableTreeNode(variable) { VariableName = variableNames[3] };
        n.AddSubtree(x3);
        n.AddSubtree(x4);
        var bb = new BuildingBlock { Name = "(* x3 x4)", Fragment = new Fragment { Root = n } };
        buildingBlockGroups.Add(new BuildingBlockGroup { Name = "(* x3 x4)", BuildingBlocks = new List<BuildingBlock> { bb } });
      }
      {
        // (* x5 x6)
        var n = multiplication.CreateTreeNode();
        var x5 = new VariableTreeNode(variable) { VariableName = variableNames[4] };
        var x6 = new VariableTreeNode(variable) { VariableName = variableNames[5] };
        n.AddSubtree(x5);
        n.AddSubtree(x6);
        var bb = new BuildingBlock { Name = "(* x5 x6)", Fragment = new Fragment { Root = n } };
        buildingBlockGroups.Add(new BuildingBlockGroup { Name = "(* x5 x6)", BuildingBlocks = new List<BuildingBlock> { bb } });
      }
      {
        var group = new BuildingBlockGroup { Name = "(* x1 x7 x9)", BuildingBlocks = new List<BuildingBlock>() };
        {
          // (* x1 x7 x9) => (* (* x7 x9) x1)
          var m1 = multiplication.CreateTreeNode();
          var m2 = multiplication.CreateTreeNode();
          var x1 = new VariableTreeNode(variable) { VariableName = variableNames[0] };
          var x7 = new VariableTreeNode(variable) { VariableName = variableNames[6] };
          var x9 = new VariableTreeNode(variable) { VariableName = variableNames[8] };
          m2.AddSubtree(x7);
          m2.AddSubtree(x9);
          m1.AddSubtree(m2);
          m1.AddSubtree(x1);
          var bb = new BuildingBlock { Name = "(* (* x7 x9) x1)", Fragment = new Fragment { Root = m1 } };
          group.BuildingBlocks.Add(bb);
        }
        {
          // (* x1 x7 x9) => (* (* x1 x9) x7)
          var m1 = multiplication.CreateTreeNode();
          var m2 = multiplication.CreateTreeNode();
          var x1 = new VariableTreeNode(variable) { VariableName = variableNames[0] };
          var x7 = new VariableTreeNode(variable) { VariableName = variableNames[6] };
          var x9 = new VariableTreeNode(variable) { VariableName = variableNames[8] };
          m2.AddSubtree(x1);
          m2.AddSubtree(x9);
          m1.AddSubtree(m2);
          m1.AddSubtree(x7);
          var bb = new BuildingBlock { Name = "(* (* x1 x9) x7)", Fragment = new Fragment { Root = m1 } };
          group.BuildingBlocks.Add(bb);
        }
        {  // (* x1 x7 x9) => (* (* x1 x7) x9)
          var m1 = multiplication.CreateTreeNode();
          var m2 = multiplication.CreateTreeNode();
          var x1 = new VariableTreeNode(variable) { VariableName = variableNames[0] };
          var x7 = new VariableTreeNode(variable) { VariableName = variableNames[6] };
          var x9 = new VariableTreeNode(variable) { VariableName = variableNames[8] };
          m2.AddSubtree(x1);
          m2.AddSubtree(x7);
          m1.AddSubtree(m2);
          m1.AddSubtree(x9);
          var bb = new BuildingBlock { Name = "(* (* x1 x7) x9)", Fragment = new Fragment { Root = m1 } };
          group.BuildingBlocks.Add(bb);
        }
        buildingBlockGroups.Add(group);
      }
      {
        var group = new BuildingBlockGroup { Name = "(* x3 x6 x10)", BuildingBlocks = new List<BuildingBlock>() };
        {
          // (* x3 x6 x10) => (* (* x6 x10) x3)
          var m1 = multiplication.CreateTreeNode();
          var m2 = multiplication.CreateTreeNode();
          var x3 = new VariableTreeNode(variable) { VariableName = variableNames[2] };
          var x6 = new VariableTreeNode(variable) { VariableName = variableNames[5] };
          var x10 = new VariableTreeNode(variable) { VariableName = variableNames[9] };
          m2.AddSubtree(x6);
          m2.AddSubtree(x10);
          m1.AddSubtree(m2);
          m1.AddSubtree(x3);
          var bb = new BuildingBlock { Name = "(* (* x6 x10) x3)", Fragment = new Fragment { Root = m1 } };
          group.BuildingBlocks.Add(bb);
        }
        {
          // (* x3 x6 x10) => (* (* x3 x10) x6)
          var m1 = multiplication.CreateTreeNode();
          var m2 = multiplication.CreateTreeNode();
          var x3 = new VariableTreeNode(variable) { VariableName = variableNames[2] };
          var x6 = new VariableTreeNode(variable) { VariableName = variableNames[5] };
          var x10 = new VariableTreeNode(variable) { VariableName = variableNames[9] };
          m2.AddSubtree(x3);
          m2.AddSubtree(x10);
          m1.AddSubtree(m2);
          m1.AddSubtree(x6);
          var bb = new BuildingBlock { Name = "(* (* x3 x10) x6)", Fragment = new Fragment { Root = m1 } };
          group.BuildingBlocks.Add(bb);
        }
        {
          // (* x3 x6 x10) => (* (* x3 x6) x10)
          var m1 = multiplication.CreateTreeNode();
          var m2 = multiplication.CreateTreeNode();
          var x3 = new VariableTreeNode(variable) { VariableName = variableNames[2] };
          var x6 = new VariableTreeNode(variable) { VariableName = variableNames[5] };
          var x10 = new VariableTreeNode(variable) { VariableName = variableNames[9] };
          m2.AddSubtree(x3);
          m2.AddSubtree(x6);
          m1.AddSubtree(m2);
          m1.AddSubtree(x10);
          var bb = new BuildingBlock { Name = "(* (* x3 x6) x10)", Fragment = new Fragment { Root = m1 } };
          group.BuildingBlocks.Add(bb);
        }
        buildingBlockGroups.Add(group);
      }
    }
  }
  internal class BuildingBlockGroup {
    public string Name;
    public List<BuildingBlock> BuildingBlocks;
  }

  internal class BuildingBlock {
    public string Name;
    public IFragment Fragment;
  }
}
